/** @module : space_agent_core
 *  @author : Adaptive & Secure Computing Systems (ASCS) Laboratory

 *  Copyright (c) 2018 ABAQS (ASCS/ECE/BU)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

/* This module reads the chemical concentration of neighboring modules and
 * computes the next concentration. Then it outputs it to the FIFOs so other
 * modules can read it.
 */

module space_agent_core #(
  parameter BIT_WIDTH = 32,
  parameter START     = {BIT_WIDTH{1'b0}} // Initial value of chemical field after reset_n
)(
  input clock,
  input reset_n,

  input [3:0] empty_from_neighbor,     // get status from neighbors
  input empty_from_cell,               // empty signal for the cell to space fifo
  input [BIT_WIDTH-1:0] north_in,
  input [BIT_WIDTH-1:0] south_in,
  input [BIT_WIDTH-1:0] east_in,
  input [BIT_WIDTH-1:0] west_in,
  input [BIT_WIDTH-1:0] cell_in,
  input [3:0] empty_from_local,        // send status to neighbors
  input empty_from_local_cell,         // empty signal for the space to cell fifo
  output reg [3:0] write_rq_out,       // send write requests to local fifo
  output reg [3:0] read_rq_out,        // send read requests to neighbors
  output reg cell_write_rq_out,
  output reg cell_read_rq_out,
  output [BIT_WIDTH-1:0] data_out,

                                       // Data Shift Register
  input write_shift_reg,
  input read_shift_reg,
  input [BIT_WIDTH-1:0] shift_in,
  output reg [BIT_WIDTH-1:0] shift_out

);


reg [3:0] nsew_new_data;   // signals if data has been writen to neighbor yet
reg cell_new_data;         // signals if data has been writen to cell agent yet
reg [BIT_WIDTH-1:0] value;

wire [BIT_WIDTH-1:0] dv;
wire [BIT_WIDTH-1:0] dvdt;
wire [BIT_WIDTH-1:0] ns_sum;
wire [BIT_WIDTH-1:0] ew_sum;
wire [BIT_WIDTH-1:0] neighbor_sum;
wire [BIT_WIDTH-1:0] difference;

//      ((south+north + west + east) - 4 * value)
// dv = -----------------------------------------
//                (4 *  delta_x^2)

// For now assume delta_x is 1
assign ns_sum       = north_in + south_in;
assign ew_sum       = east_in  + west_in;
assign neighbor_sum = ns_sum  + ew_sum;
assign difference   = neighbor_sum - {value[29:0],2'b00};
assign dv           = { {2{difference[31]}}, difference[31:2]};
// TODO: Parameterize dt, constrain to power of 2
// TODO: Determine if cell_in has already been scaled by dt
//       Should it be in "assign dvdt" or "assign dv"
// dvdt = (dv*dt) + cell_in
assign dvdt         = { dv[31], dv[31:1]} + cell_in;
assign data_out     = value;


always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    value            <= START;
    read_rq_out      <= 4'h0;
    cell_read_rq_out <= 1'b0;
    nsew_new_data    <= 4'hf;
    cell_new_data    <= 1'b1;
  end else if ( read_shift_reg )begin
    // TODO: FIXME: Test initialization after loading data
    // For now, act as if the space agent just read all of the neightbors
    // This will force them to be empty, allowing neighbors to write to them
    // again.
    //
    // Must make sure that invalid old data is not read before the new data is
    // written
    value            <= shift_out;
    read_rq_out      <= 4'hf; // signal that the data has been read
    cell_read_rq_out <= 1'b1;
    nsew_new_data    <= 4'hf;
    cell_new_data    <= 1'b1;
  end else if (
                empty_from_neighbor == 4'h0 &
                empty_from_cell     == 1'b0 &
                nsew_new_data       == 4'h0 &
                cell_new_data       == 1'b0
              )
  begin
    // update when neighbors are full and all of the new data has been written
    value            <= value + dvdt;
    read_rq_out      <= 4'hf; // signal that the data has been read
    cell_read_rq_out <= 1'b1;
    nsew_new_data    <= 4'hf;
    cell_new_data    <= 1'b1;
  end else begin
    value            <= value;
    read_rq_out      <= 4'h0;
    cell_read_rq_out <= 1'b0;
    // if wr == 1 then set new data to 0
    nsew_new_data    <= nsew_new_data & ( ~write_rq_out );
    cell_new_data    <= cell_new_data & ( ~cell_write_rq_out );
  end
end

always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    write_rq_out      <= 4'h0;
    cell_write_rq_out <= 1'b0;
  end else begin
    // signal that data_out should be latched
    write_rq_out      <= empty_from_local & nsew_new_data;
    cell_write_rq_out <= empty_from_local_cell & cell_new_data;
  end
end


// Data Shift Register
always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    shift_out <= {BIT_WIDTH{1'b0}};
  end else if( write_shift_reg ) begin
    shift_out <= value;
  end else begin
    shift_out <= shift_in;
  end
end

endmodule
